#ifndef _SWELLINGTON_RFID522_SPI_H
#define _SWELLINGTON_RFID522_SPI_H

#pragma once

#include <Mitov.h>
#include <Swellington_RFID_Mifare.h>

namespace Swellington
{

    template <
        typename T_Inverted,
        typename T_OutputPin>
    class TArduinoRFID522Key : private T_Inverted,
                               private T_OutputPin
    {
    private:
        _V_PIN_(OutputPin)
        _V_PROP_(Inverted)

    public:
        bool IsConnected()
        {
            return OutputPin().GetPinIsConnected();
        }

        inline void SetSerialKey(uint32_t num)
        {
            if (Inverted())
            {
                num = (num >> 24) |
                      ((num >> 8) & 0x0000FF00) |
                      ((num << 8) & 0x00FF0000) |
                      (num << 24);
            }

            OutputPin().SetPinValue(num);
        }
    };

    template <typename T_SPI, T_SPI &C_SPI,
              typename T_ChipSelectOutputPin,
              typename T_Enabled,
              typename T_Key,
              uint8_t C_ResetOutputPin>
    class RFID522_SPI : protected T_ChipSelectOutputPin,
                        protected T_Enabled,
                        protected T_Key
    {
    protected:
        _V_PIN_(ChipSelectOutputPin)
        _V_PROP_(Enabled)
        _V_PROP_(Key)

    private:
        Mifare::Uid _uid;

    private:
        void PCD_SetRegisterBitMask(Mifare::PCD_Register reg, uint8_t mask)
        {
            uint8_t tmp = PCD_ReadRegister(reg);
            PCD_WriteRegister(reg, tmp | mask);
        }

        void PCD_ClearRegisterBitMask(Mifare::PCD_Register reg, uint8_t mask)
        {
            uint8_t temp = PCD_ReadRegister(reg);
            PCD_WriteRegister(reg, temp & (~mask));
        }

        void PCD_WriteRegister(Mifare::PCD_Register reg, uint8_t count, uint8_t *values)
        {
            C_SPI.beginTransaction(SPISettings(4000000u, MSBFIRST, SPI_MODE0));
            ChipSelectOutputPin().SetPinValueLow();
            C_SPI.transfer(reg & 0x7E);
            for (int index = 0; index < count; index++)
            {
                C_SPI.transfer(values[index]);
            }
            ChipSelectOutputPin().SetPinValueHigh();
            C_SPI.endTransaction();
        }

        void PCD_WriteRegister(Mifare::PCD_Register reg, uint8_t value)
        {
            C_SPI.beginTransaction(SPISettings(4000000u, MSBFIRST, SPI_MODE0));
            ChipSelectOutputPin().SetPinValueLow();
            C_SPI.transfer(reg);
            C_SPI.transfer(value);
            ChipSelectOutputPin().SetPinValueHigh();
            C_SPI.endTransaction();
        }

        void PCD_ReadRegister(Mifare::PCD_Register reg, uint8_t count, uint8_t *values, uint8_t rxAlign = 0)
        {
            if (count == 0)
                return;
            uint8_t address = 0x80 | (reg & 0x7E);
            uint8_t index = 0;
            C_SPI.beginTransaction(SPISettings(4000000u, MSBFIRST, SPI_MODE0));
            ChipSelectOutputPin().SetPinValueLow();
            count--;
            SPI.transfer(address);
            while (index < count)
            {
                if (index == 0 && rxAlign)
                {
                    uint8_t mask = 0;
                    for (int i = rxAlign; i <= 7; i++)
                    {
                        mask |= (1 << i);
                    }
                    uint8_t value = SPI.transfer(address);
                    values[0] = (values[index] & ~mask) | (value & mask);
                }
                else
                {
                    values[index] = SPI.transfer(address);
                }
                index++;
            }
            values[index] = SPI.transfer(0);
            ChipSelectOutputPin().SetPinValueHigh();
            C_SPI.endTransaction();
        }

        uint8_t PCD_ReadRegister(Mifare::PCD_Register reg)
        {
            uint8_t value;
            C_SPI.beginTransaction(SPISettings(4000000u, MSBFIRST, SPI_MODE0));
            ChipSelectOutputPin().SetPinValueLow();
            C_SPI.transfer(0x80 | reg);
            value = C_SPI.transfer(0);
            ChipSelectOutputPin().SetPinValueHigh();
            C_SPI.endTransaction();
            return value;
        }

        void PCD_AntennaOn()
        {
            uint8_t value = PCD_ReadRegister(Mifare::PCD_Register::TxControlReg);
            if ((value & 0x03) != 0x03)
            {
                PCD_WriteRegister(Mifare::PCD_Register::TxControlReg, value | 0x03);
            }
        }

        Mifare::StatusCode PCD_CalculateCRC(uint8_t *data,  ///< In: Pointer to the data to transfer to the FIFO for CRC calculation.
                                            uint8_t length, ///< In: The number of bytes to transfer.
                                            uint8_t *result ///< Out: Pointer to result buffer. Result is written to result[0..1], low byte first.
        )
        {
            PCD_WriteRegister(Mifare::PCD_Register::CommandReg, Mifare::PCD_Command::PCD_Idle);    // Stop any active command.
            PCD_WriteRegister(Mifare::PCD_Register::DivIrqReg, 0x04);                              // Clear the CRCIRq interrupt request bit
            PCD_WriteRegister(Mifare::PCD_Register::FIFOLevelReg, 0x80);                           // FlushBuffer = 1, FIFO initialization
            PCD_WriteRegister(Mifare::PCD_Register::FIFODataReg, length, data);                    // Write data to the FIFO
            PCD_WriteRegister(Mifare::PCD_Register::CommandReg, Mifare::PCD_Command::PCD_CalcCRC); // Start the calculation

            // Wait for the CRC calculation to complete. Check for the register to
            // indicate that the CRC calculation is complete in a loop. If the
            // calculation is not indicated as complete in ~90ms, then time out
            // the operation.
            const uint32_t deadline = millis() + 89;

            do
            {
                // DivIrqReg[7..0] bits are: Set2 reserved reserved MfinActIRq reserved CRCIRq reserved reserved
                uint8_t n = PCD_ReadRegister(Mifare::PCD_Register::DivIrqReg);
                if (n & 0x04)
                {                                                                                       // CRCIRq bit set - calculation done
                    PCD_WriteRegister(Mifare::PCD_Register::CommandReg, Mifare::PCD_Command::PCD_Idle); // Stop calculating CRC for new content in the FIFO.
                    // Transfer the result from the registers to the result buffer
                    result[0] = PCD_ReadRegister(Mifare::PCD_Register::CRCResultRegL);
                    result[1] = PCD_ReadRegister(Mifare::PCD_Register::CRCResultRegH);
                    return Mifare::StatusCode::STATUS_OK;
                }
                yield();
            } while (static_cast<uint32_t>(millis()) < deadline);

            // 89ms passed and nothing happened. Communication with the MFRC522 might be down.
            return Mifare::StatusCode::STATUS_TIMEOUT;
        }

        Mifare::StatusCode PCD_TransceiveData(uint8_t *sendData,            ///< Pointer to the data to transfer to the FIFO.
                                              uint8_t sendLen,              ///< Number of bytes to transfer to the FIFO.
                                              uint8_t *backData,            ///< nullptr or pointer to buffer if data should be read back after executing the command.
                                              uint8_t *backLen,             ///< In: Max number of bytes to write to *backData. Out: The number of bytes returned.
                                              uint8_t *validBits = nullptr, ///< In/Out: The number of valid bits in the last byte. 0 for 8 valid bits. Default nullptr.
                                              uint8_t rxAlign = 0,          ///< In: Defines the bit position in backData[0] for the first bit received. Default 0.
                                              bool checkCRC = false         ///< In: True => The last two bytes of the response is assumed to be a CRC_A that must be validated.
        )
        {
            uint8_t waitIRq = 0x30; // RxIRq and IdleIRq
            return PCD_CommunicateWithPICC(Mifare::PCD_Command::PCD_Transceive, waitIRq, sendData, sendLen, backData, backLen, validBits, rxAlign, checkCRC);
        }

        Mifare::StatusCode PCD_CommunicateWithPICC(Mifare::PCD_Command command, ///< The command to execute. One of the PCD_Command enums.
                                                   uint8_t waitIRq,             ///< The bits in the ComIrqReg register that signals successful completion of the command.
                                                   uint8_t *sendData,           ///< Pointer to the data to transfer to the FIFO.
                                                   uint8_t sendLen,             ///< Number of bytes to transfer to the FIFO.
                                                   uint8_t *backData,           ///< nullptr or pointer to buffer if data should be read back after executing the command.
                                                   uint8_t *backLen,            ///< In: Max number of bytes to write to *backData. Out: The number of bytes returned.
                                                   uint8_t *validBits,          ///< In/Out: The number of valid bits in the last byte. 0 for 8 valid bits.
                                                   uint8_t rxAlign,             ///< In: Defines the bit position in backData[0] for the first bit received. Default 0.
                                                   bool checkCRC                ///< In: True => The last two bytes of the response is assumed to be a CRC_A that must be validated.
        )
        {
            // Prepare values for BitFramingReg
            byte txLastBits = validBits ? *validBits : 0;
            byte bitFraming = (rxAlign << 4) + txLastBits; // RxAlign = BitFramingReg[6..4]. TxLastBits = BitFramingReg[2..0]

            PCD_WriteRegister(Mifare::PCD_Register::CommandReg, Mifare::PCD_Command::PCD_Idle); // Stop any active command.
            PCD_WriteRegister(Mifare::PCD_Register::ComIrqReg, 0x7F);                           // Clear all seven interrupt request bits
            PCD_WriteRegister(Mifare::PCD_Register::FIFOLevelReg, 0x80);                        // FlushBuffer = 1, FIFO initialization
            PCD_WriteRegister(Mifare::PCD_Register::FIFODataReg, sendLen, sendData);            // Write sendData to the FIFO
            PCD_WriteRegister(Mifare::PCD_Register::BitFramingReg, bitFraming);                 // Bit adjustments
            PCD_WriteRegister(Mifare::PCD_Register::CommandReg, command);                       // Execute the command
            if (command == Mifare::PCD_Command::PCD_Transceive)
            {
                PCD_SetRegisterBitMask(Mifare::PCD_Register::BitFramingReg, 0x80); // StartSend=1, transmission of data starts
            }

            // In PCD_Init() we set the TAuto flag in TModeReg. This means the timer
            // automatically starts when the PCD stops transmitting.
            //
            // Wait here for the command to complete. The bits specified in the
            // `waitIRq` parameter define what bits constitute a completed command.
            // When they are set in the ComIrqReg register, then the command is
            // considered complete. If the command is not indicated as complete in
            // ~36ms, then consider the command as timed out.
            const uint32_t deadline = millis() + 36;
            bool completed = false;

            do
            {
                byte n = PCD_ReadRegister(Mifare::PCD_Register::ComIrqReg); // ComIrqReg[7..0] bits are: Set1 TxIRq RxIRq IdleIRq HiAlertIRq LoAlertIRq ErrIRq TimerIRq
                if (n & waitIRq)
                { // One of the interrupts that signal success has been set.
                    completed = true;
                    break;
                }
                if (n & 0x01)
                { // Timer interrupt - nothing received in 25ms
                    return Mifare::StatusCode::STATUS_TIMEOUT;
                }
                yield();
            } while (static_cast<uint32_t>(millis()) < deadline);

            // 36ms and nothing happened. Communication with the MFRC522 might be down.
            if (!completed)
            {
                return Mifare::StatusCode::STATUS_TIMEOUT;
            }

            // Stop now if any errors except collisions were detected.
            byte errorRegValue = PCD_ReadRegister(Mifare::PCD_Register::ErrorReg); // ErrorReg[7..0] bits are: WrErr TempErr reserved BufferOvfl CollErr CRCErr ParityErr ProtocolErr
            if (errorRegValue & 0x13)
            { // BufferOvfl ParityErr ProtocolErr
                return Mifare::StatusCode::STATUS_ERROR;
            }

            uint8_t _validBits = 0;

            // If the caller wants data back, get it from the MFRC522.
            if (backData && backLen)
            {
                uint8_t n = PCD_ReadRegister(Mifare::PCD_Register::FIFOLevelReg); // Number of bytes in the FIFO
                if (n > *backLen)
                {
                    return Mifare::StatusCode::STATUS_NO_ROOM;
                }
                *backLen = n;                                                              // Number of bytes returned
                PCD_ReadRegister(Mifare::PCD_Register::FIFODataReg, n, backData, rxAlign); // Get received data from FIFO
                _validBits = PCD_ReadRegister(Mifare::PCD_Register::ControlReg) & 0x07;    // RxLastBits[2:0] indicates the number of valid bits in the last received byte. If this value is 000b, the whole byte is valid.
                if (validBits)
                {
                    *validBits = _validBits;
                }
            }

            // Tell about collisions
            if (errorRegValue & 0x08)
            { // CollErr
                return Mifare::StatusCode::STATUS_COLLISION;
            }

            // Perform CRC_A validation if requested.
            if (backData && backLen && checkCRC)
            {
                // In this case a MIFARE Classic NAK is not OK.
                if (*backLen == 1 && _validBits == 4)
                {
                    return Mifare::StatusCode::STATUS_MIFARE_NACK;
                }
                // We need at least the CRC_A value and all 8 bits of the last byte must be received.
                if (*backLen < 2 || _validBits != 0)
                {
                    return Mifare::StatusCode::STATUS_CRC_WRONG;
                }
                // Verify CRC_A - do our own calculation and store the control in controlBuffer.
                byte controlBuffer[2];
                Mifare::StatusCode status = PCD_CalculateCRC(&backData[0], *backLen - 2, &controlBuffer[0]);
                if (status != Mifare::StatusCode::STATUS_OK)
                {
                    return status;
                }
                if ((backData[*backLen - 2] != controlBuffer[0]) || (backData[*backLen - 1] != controlBuffer[1]))
                {
                    return Mifare::StatusCode::STATUS_CRC_WRONG;
                }
            }

            return Mifare::StatusCode::STATUS_OK;
        }

        Mifare::StatusCode PICC_REQA_or_WUPA(uint8_t command, uint8_t *bufferATQA, uint8_t *bufferSize)
        {
            uint8_t validBits;
            Mifare::StatusCode status;
            if (bufferATQA == nullptr || *bufferSize < 2)
                return Mifare::StatusCode::STATUS_NO_ROOM;
            PCD_ClearRegisterBitMask(Mifare::PCD_Register::CollReg, 0x80);
            validBits = 7;
            status = PCD_TransceiveData(&command, 1, bufferATQA, bufferSize, &validBits);
            if (status != Mifare::StatusCode::STATUS_OK)
            {
                return status;
            }
            if (*bufferSize != 2 || validBits != 0)
            { // ATQA must be exactly 16 bits.
                return Mifare::StatusCode::STATUS_ERROR;
            }
            return Mifare::StatusCode::STATUS_OK;
        }

        Mifare::StatusCode PICC_RequestA(uint8_t *bufferATQA, uint8_t *bufferSize)
        {
            return PICC_REQA_or_WUPA(Mifare::PICC_Command::PICC_CMD_REQA, bufferATQA, bufferSize);
        }

        Mifare::StatusCode PICC_Select(Mifare::Uid *uid, uint8_t validBits = 0)
        {
            bool uidComplete;
            bool selectDone;
            bool useCascadeTag;
            byte cascadeLevel = 1;
            Mifare::StatusCode result;
            byte count;
            byte checkBit;
            byte index;
            byte uidIndex;                // The first index in uid->uidByte[] that is used in the current Cascade Level.
            int8_t currentLevelKnownBits; // The number of known UID bits in the current Cascade Level.
            byte buffer[9];               // The SELECT/ANTICOLLISION commands uses a 7 byte standard frame + 2 bytes CRC_A
            byte bufferUsed;              // The number of bytes used in the buffer, ie the number of bytes to transfer to the FIFO.
            byte rxAlign;                 // Used in BitFramingReg. Defines the bit position for the first bit received.
            byte txLastBits;              // Used in BitFramingReg. The number of valid bits in the last transmitted byte.
            byte *responseBuffer;
            byte responseLength;

            // Sanity checks
            if (validBits > 80)
            {
                return Mifare::StatusCode::STATUS_INVALID;
            }

            // Prepare MFRC522
            PCD_ClearRegisterBitMask(Mifare::PCD_Register::CollReg, 0x80); // ValuesAfterColl=1 => Bits received after collision are cleared.

            // Repeat Cascade Level loop until we have a complete UID.
            uidComplete = false;
            while (!uidComplete)
            {
                // Set the Cascade Level in the SEL byte, find out if we need to use the Cascade Tag in byte 2.
                switch (cascadeLevel)
                {
                case 1:
                    buffer[0] = Mifare::PICC_Command::PICC_CMD_SEL_CL1;
                    uidIndex = 0;
                    useCascadeTag = validBits && uid->size > 4; // When we know that the UID has more than 4 bytes
                    break;

                case 2:
                    buffer[0] = Mifare::PICC_Command::PICC_CMD_SEL_CL2;
                    uidIndex = 3;
                    useCascadeTag = validBits && uid->size > 7; // When we know that the UID has more than 7 bytes
                    break;

                case 3:
                    buffer[0] = Mifare::PICC_Command::PICC_CMD_SEL_CL3;
                    uidIndex = 6;
                    useCascadeTag = false; // Never used in CL3.
                    break;

                default:
                    return Mifare::StatusCode::STATUS_INTERNAL_ERROR;
                    break;
                }

                // How many UID bits are known in this Cascade Level?
                currentLevelKnownBits = validBits - (8 * uidIndex);
                if (currentLevelKnownBits < 0)
                {
                    currentLevelKnownBits = 0;
                }
                // Copy the known bits from uid->uidByte[] to buffer[]
                index = 2; // destination index in buffer[]
                if (useCascadeTag)
                {
                    buffer[index++] = Mifare::PICC_Command::PICC_CMD_CT;
                }
                byte bytesToCopy = currentLevelKnownBits / 8 + (currentLevelKnownBits % 8 ? 1 : 0); // The number of bytes needed to represent the known bits for this level.
                if (bytesToCopy)
                {
                    byte maxBytes = useCascadeTag ? 3 : 4; // Max 4 bytes in each Cascade Level. Only 3 left if we use the Cascade Tag
                    if (bytesToCopy > maxBytes)
                    {
                        bytesToCopy = maxBytes;
                    }
                    for (count = 0; count < bytesToCopy; count++)
                    {
                        buffer[index++] = uid->uidByte[uidIndex + count];
                    }
                }
                // Now that the data has been copied we need to include the 8 bits in CT in currentLevelKnownBits
                if (useCascadeTag)
                {
                    currentLevelKnownBits += 8;
                }

                // Repeat anti collision loop until we can transmit all UID bits + BCC and receive a SAK - max 32 iterations.
                selectDone = false;
                while (!selectDone)
                {
                    // Find out how many bits and bytes to send and receive.
                    if (currentLevelKnownBits >= 32)
                    { // All UID bits in this Cascade Level are known. This is a SELECT.
                        // Serial.print(F("SELECT: currentLevelKnownBits=")); Serial.println(currentLevelKnownBits, DEC);
                        buffer[1] = 0x70; // NVB - Number of Valid Bits: Seven whole bytes
                        // Calculate BCC - Block Check Character
                        buffer[6] = buffer[2] ^ buffer[3] ^ buffer[4] ^ buffer[5];
                        // Calculate CRC_A
                        result = PCD_CalculateCRC(buffer, 7, &buffer[7]);
                        if (result != Mifare::StatusCode::STATUS_OK)
                        {
                            return result;
                        }
                        txLastBits = 0; // 0 => All 8 bits are valid.
                        bufferUsed = 9;
                        // Store response in the last 3 bytes of buffer (BCC and CRC_A - not needed after tx)
                        responseBuffer = &buffer[6];
                        responseLength = 3;
                    }
                    else
                    { // This is an ANTICOLLISION.
                        // Serial.print(F("ANTICOLLISION: currentLevelKnownBits=")); Serial.println(currentLevelKnownBits, DEC);
                        txLastBits = currentLevelKnownBits % 8;
                        count = currentLevelKnownBits / 8;     // Number of whole bytes in the UID part.
                        index = 2 + count;                     // Number of whole bytes: SEL + NVB + UIDs
                        buffer[1] = (index << 4) + txLastBits; // NVB - Number of Valid Bits
                        bufferUsed = index + (txLastBits ? 1 : 0);
                        // Store response in the unused part of buffer
                        responseBuffer = &buffer[index];
                        responseLength = sizeof(buffer) - index;
                    }

                    // Set bit adjustments
                    rxAlign = txLastBits;                                                                // Having a separate variable is overkill. But it makes the next line easier to read.
                    PCD_WriteRegister(Mifare::PCD_Register::BitFramingReg, (rxAlign << 4) + txLastBits); // RxAlign = BitFramingReg[6..4]. TxLastBits = BitFramingReg[2..0]

                    // Transmit the buffer and receive the response.
                    result = PCD_TransceiveData(buffer, bufferUsed, responseBuffer, &responseLength, &txLastBits, rxAlign);
                    if (result == Mifare::StatusCode::STATUS_COLLISION)
                    {                                                                          // More than one PICC in the field => collision.
                        byte valueOfCollReg = PCD_ReadRegister(Mifare::PCD_Register::CollReg); // CollReg[7..0] bits are: ValuesAfterColl reserved CollPosNotValid CollPos[4:0]
                        if (valueOfCollReg & 0x20)
                        {                                                // CollPosNotValid
                            return Mifare::StatusCode::STATUS_COLLISION; // Without a valid collision position we cannot continue
                        }
                        byte collisionPos = valueOfCollReg & 0x1F; // Values 0-31, 0 means bit 32.
                        if (collisionPos == 0)
                        {
                            collisionPos = 32;
                        }
                        if (collisionPos <= currentLevelKnownBits)
                        { // No progress - should not happen
                            return Mifare::StatusCode::STATUS_INTERNAL_ERROR;
                        }
                        // Choose the PICC with the bit set.
                        currentLevelKnownBits = collisionPos;
                        count = currentLevelKnownBits % 8; // The bit to modify
                        checkBit = (currentLevelKnownBits - 1) % 8;
                        index = 1 + (currentLevelKnownBits / 8) + (count ? 1 : 0); // First byte is index 0.
                        buffer[index] |= (1 << checkBit);
                    }
                    else if (result != Mifare::StatusCode::STATUS_OK)
                    {
                        return result;
                    }
                    else
                    { // STATUS_OK
                        if (currentLevelKnownBits >= 32)
                        {                      // This was a SELECT.
                            selectDone = true; // No more anticollision
                                               // We continue below outside the while.
                        }
                        else
                        { // This was an ANTICOLLISION.
                            // We now have all 32 bits of the UID in this Cascade Level
                            currentLevelKnownBits = 32;
                            // Run loop again to do the SELECT.
                        }
                    }
                } // End of while (!selectDone)

                // We do not check the CBB - it was constructed by us above.

                // Copy the found UID bytes from buffer[] to uid->uidByte[]
                index = (buffer[2] == Mifare::PICC_Command::PICC_CMD_CT) ? 3 : 2; // source index in buffer[]
                bytesToCopy = (buffer[2] == Mifare::PICC_Command::PICC_CMD_CT) ? 3 : 4;
                for (count = 0; count < bytesToCopy; count++)
                {
                    uid->uidByte[uidIndex + count] = buffer[index++];
                }

                // Check response SAK (Select Acknowledge)
                if (responseLength != 3 || txLastBits != 0)
                { // SAK must be exactly 24 bits (1 byte + CRC_A).
                    return Mifare::StatusCode::STATUS_ERROR;
                }
                // Verify CRC_A - do our own calculation and store the control in buffer[2..3] - those bytes are not needed anymore.
                result = PCD_CalculateCRC(responseBuffer, 1, &buffer[2]);
                if (result != Mifare::StatusCode::STATUS_OK)
                {
                    return result;
                }
                if ((buffer[2] != responseBuffer[1]) || (buffer[3] != responseBuffer[2]))
                {
                    return Mifare::StatusCode::STATUS_CRC_WRONG;
                }
                if (responseBuffer[0] & 0x04)
                { // Cascade bit set - UID not complete yes
                    cascadeLevel++;
                }
                else
                {
                    uidComplete = true;
                    uid->sak = responseBuffer[0];
                }
            } // End of while (!uidComplete)

            // Set correct uid->size
            uid->size = 3 * cascadeLevel + 1;

            return Mifare::StatusCode::STATUS_OK;
        }

        bool PICC_ReadCardSerial()
        {
            Mifare::StatusCode status = PICC_Select(&_uid);
            return status == Mifare::StatusCode::STATUS_OK;
        }

        bool PICC_IsNewCardPresent()
        {
            byte bufferATQA[2];
            byte bufferSize = sizeof(bufferATQA);

            // Reset baud rates
            PCD_WriteRegister(Mifare::PCD_Register::TxModeReg, 0x00);
            PCD_WriteRegister(Mifare::PCD_Register::RxModeReg, 0x00);
            // Reset ModWidthReg
            PCD_WriteRegister(Mifare::PCD_Register::ModWidthReg, 0x26);

            Mifare::StatusCode result = PICC_RequestA(bufferATQA, &bufferSize);

            return (result == Mifare::StatusCode::STATUS_OK || result == Mifare::StatusCode::STATUS_COLLISION);
        }

        Mifare::StatusCode PICC_HaltA()
        {
            Mifare::StatusCode result;
            uint8_t buffer[4];

            // Build command buffer
            buffer[0] = Mifare::PICC_Command::PICC_CMD_HLTA;
            buffer[1] = 0;
            // Calculate CRC_A
            result = PCD_CalculateCRC(buffer, 2, &buffer[2]);
            if (result != Mifare::StatusCode::STATUS_OK)
            {
                return result;
            }

            result = PCD_TransceiveData(buffer, sizeof(buffer), NULL, 0);
            if (result == Mifare::StatusCode::STATUS_TIMEOUT)
            {
                return Mifare::StatusCode::STATUS_OK;
            }
            if (result == Mifare::StatusCode::STATUS_OK)
            { // That is ironically NOT ok in this case ;-)
                return Mifare::StatusCode::STATUS_ERROR;
            }
            return result;
        }

        void PCD_StopCrypto1()
        {
            PCD_ClearRegisterBitMask(Mifare::PCD_Register::Status2Reg, 0x08);
        }

        void Soft_Reset()
        {
            PCD_WriteRegister(Mifare::PCD_Register::CommandReg, Mifare::PCD_Command::PCD_SoftReset);
            uint8_t count = 0;
            do
            {
                delay(50);
            } while ((PCD_ReadRegister(Mifare::PCD_Register::CommandReg) & (1 << 4)) && (++count) < 3);
        }

    public:
        inline void ClockInputPin_o_Receive(void *)
        {
            if (!Enabled() || (!Key().IsConnected()))
            {
                return;
            }

            if (PICC_IsNewCardPresent() && PICC_ReadCardSerial())
            {
                uint32_t k = 0;
                memcpy(&k, _uid.uidByte, _uid.size);
                Key().SetSerialKey(k);
                PICC_HaltA();
            }
        }

        inline void SystemInit()
        {
            if (!Enabled())
                return;
            bool hardReset = false;
            ChipSelectOutputPin().SetPinValueHigh();
            if (C_ResetOutputPin != UINT8_MAX)
            {
                pinMode(C_ResetOutputPin, INPUT);
                if (digitalRead(C_ResetOutputPin) == LOW)
                {
                    pinMode(C_ResetOutputPin, OUTPUT);
                    digitalWrite(C_ResetOutputPin, LOW);
                    delayMicroseconds(2);
                    digitalWrite(C_ResetOutputPin, HIGH);
                    delay(50);
                    hardReset = true;
                }
            }

            if (!hardReset)
            {
                Soft_Reset();
            }

            PCD_WriteRegister(Mifare::PCD_Register::TxModeReg, 0x00);
            PCD_WriteRegister(Mifare::PCD_Register::RxModeReg, 0x00);
            PCD_WriteRegister(Mifare::PCD_Register::ModWidthReg, 0x26);
            PCD_WriteRegister(Mifare::PCD_Register::TModeReg, 0x80);
            PCD_WriteRegister(Mifare::PCD_Register::TPrescalerReg, 0xA9);
            PCD_WriteRegister(Mifare::PCD_Register::TReloadRegH, 0x03);
            PCD_WriteRegister(Mifare::PCD_Register::TReloadRegL, 0xE8);
            PCD_WriteRegister(Mifare::PCD_Register::TxASKReg, 0x40);
            PCD_WriteRegister(Mifare::PCD_Register::ModeReg, 0x3D);
            PCD_AntennaOn();
            delay(1000);
        }

        inline void SystemLoopBegin()
        {
            ClockInputPin_o_Receive(nullptr);
        }
    };
}

#endif